#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#ifdef TURBOC
#include <dos.h>
#include <alloc.h>
#include <graphics.h>
#endif
#include "wirefram.h"

#define MAX_CHOP_PIECES 32
#ifndef TURBOC
#define WHITE 255
#endif

extern int Shade_Flag;
extern double offsetx, offsety;
extern int window_width, window_height;

extern OBJECT *All_Objects;
extern OBJECT *Current_Display_Object;
extern VIEWPOINT *Eye;
extern TRANSFORMATION WS;
extern DBL VTemp;

void
world_position_object(Object, parametric_coordinate, world_coordinate)
   OBJECT *Object;
   VECTOR *parametric_coordinate, *world_coordinate;
{
   DBL x, y, z;
   VECTOR v0, v1;

   x = Object->Orientation.Scale.x * parametric_coordinate->x;
   y = Object->Orientation.Scale.y * parametric_coordinate->y;
   z = Object->Orientation.Scale.z * parametric_coordinate->z;

   v0 = Object->Orientation.Origin;
   VScale(v1, Object->Orientation.Major_Axis, z);
   VAdd(v0, v0, v1);
   VScale(v1, Object->Orientation.Minor_Axis1, y);
   VAdd(v0, v0, v1);
   VScale(v1, Object->Orientation.Minor_Axis2, x);
   VAdd(v0, v0, v1);
   *world_coordinate = v0;
}

void
world_position_normal(Object, parametric_coordinate, world_coordinate)
   OBJECT *Object;
   VECTOR *parametric_coordinate, *world_coordinate;
{
   DBL x, y, z;
   VECTOR v0, v1;

   x = Object->Orientation.Scale.x * parametric_coordinate->x;
   y = Object->Orientation.Scale.y * parametric_coordinate->y;
   z = Object->Orientation.Scale.z * parametric_coordinate->z;

   VScale(v0, Object->Orientation.Major_Axis, z);
   VAdd(v0, v0, v1);
   VScale(v1, Object->Orientation.Minor_Axis1, y);
   VAdd(v0, v0, v1);
   VScale(v1, Object->Orientation.Minor_Axis2, x);
   VAdd(v0, v0, v1);
   *world_coordinate = v0;
}

void
reset_viewpoint()
{
   VSet(Eye->Orientation.Origin, 0.0, 0.0, -10.0);
   VSet(Eye->Orientation.Major_Axis,  0.0, 0.0, 1.0);
   VSet(Eye->Orientation.Minor_Axis1, 0.0, 1.0, 0.0);
   VSet(Eye->Orientation.Minor_Axis2, 1.0, 0.0, 0.0);
   VSet(Eye->Orientation.Scale, 1.0, 1.0, 1.0);
   VSet(Eye->Focus->Look_At, 0.0, 0.0, 0.0);
   Eye->Focus->Distance = 10.0;
   Eye->Focus->Field_Of_View = atan(0.5);
   Normalize_View(&WS);
}

int
degenerate_line(vert0, vert1)
   VECTOR *vert0, *vert1;
{
   if (fabs(vert0->x - vert1->x) < EPSILON &&
       fabs(vert0->y - vert1->y) < EPSILON &&
       fabs(vert0->z - vert1->z) < EPSILON)
      return 1;
   else
      return 0;
}

void
Generic_Displayer(Object, color)
   OBJECT *Object;
   TEXTURE *color;
{
   long offset;
   int i, j, k;
   DBL xp[2], yp[2];
   DBL u, v, x, y, z, param[2], *param_ptr;
   VECTOR v0, v1;
   DBL u_min = Object->Parameters->Min_U;
   DBL u_max = Object->Parameters->Max_U;
   DBL v_min = Object->Parameters->Min_V;
   DBL v_max = Object->Parameters->Max_V;
   int u_steps = Object->Parameters->U_Steps;
   int v_steps = Object->Parameters->V_Steps;
   DBL u_range, v_range, delta_u, delta_v;
   VECTOR vertices[4];
   DBL plane[4];
   int strip_count, l;
   Poly Polygon;

   Current_Display_Object = Object;

   param_ptr = &param[0];

   u_range = u_max - u_min;
   v_range = v_max - v_min;

   /* Calculate the grid and project it. */
   delta_u = u_range / (DBL)u_steps;
   delta_v = v_range / (DBL)v_steps;

   if (!Shade_Flag) {
      for (i=0,u=u_min;i<=u_steps;i++,u+=delta_u) {
         for (j=0,v=v_min;j<=v_steps;j++,v+=delta_v) {
            param[0] = u; param[1] = v;
            Evaluate(Object, param_ptr, &v0);
            world_position_object(Object, &v0, &vertices[1]);
   
            if (j > 0) {
               if (degenerate_line(&vertices[0], &vertices[1]))
                  continue;
               project_points(Eye, 2, &vertices[0], &xp[0], &yp[0], NULL);
               show_polygon(color, 1, 2, &xp[0], &yp[0]);
               }
            vertices[0] = vertices[1];
            }
         }

      for (j=0,v=v_min;j<=v_steps;j++,v+=delta_v) {
         for (i=0,u=u_min;i<=u_steps;i++,u+=delta_u) {
            param[0] = u; param[1] = v;
            Evaluate(Object, param_ptr, &v0);
            world_position_object(Object, &v0, &vertices[1]);
            if (i > 0) {
               project_points(Eye, 2, &vertices[0], &xp[0], &yp[0], NULL);
               show_polygon(color, 1, 2, &xp[0], &yp[0]);
               }
            vertices[0] = vertices[1];
            }
         }
      }
   else {
      /* Apply lighting to each polygon. */
      for (i=0,u=u_min;i<u_steps;i++,u+=delta_u) {
         for (j=0,v=v_min;j<v_steps;j++,v+=delta_v) {
            param[0] = u; param[1] = v;
            Evaluate(Object, param_ptr, &v0);
            Normal(Object, param_ptr, &v1);
            world_position_object(Object, &v0, &v0);
            world_position_normal(Object, &v1, &v1);
            VNormalize(v1, v1);
            Polygon.vert[0].x  = v0.x;
            Polygon.vert[0].y  = v0.y;
            Polygon.vert[0].z  = v0.z;
            Polygon.vert[0].nx = v1.x;
            Polygon.vert[0].ny = v1.y;
            Polygon.vert[0].nz = v1.z;

            param[1] = v+delta_v;
            Evaluate(Object, param_ptr, &v0);
            Normal(Object, param_ptr, &v1);
            world_position_object(Object, &v0, &v0);
            world_position_normal(Object, &v1, &v1);
            VNormalize(v1, v1);
            Polygon.vert[1].x  = v0.x;
            Polygon.vert[1].y  = v0.y;
            Polygon.vert[1].z  = v0.z;
            Polygon.vert[1].nx = v1.x;
            Polygon.vert[1].ny = v1.y;
            Polygon.vert[1].nz = v1.z;

            param[0] = u+delta_u;
            Evaluate(Object, param_ptr, &v0);
            Normal(Object, param_ptr, &v1);
            world_position_object(Object, &v0, &v0);
            world_position_normal(Object, &v1, &v1);
            VNormalize(v1, v1);
            Polygon.vert[2].x  = v0.x;
            Polygon.vert[2].y  = v0.y;
            Polygon.vert[2].z  = v0.z;
            Polygon.vert[2].nx = v1.x;
            Polygon.vert[2].ny = v1.y;
            Polygon.vert[2].nz = v1.z;

            param[1] = v;
            Evaluate(Object, param_ptr, &v0);
            Normal(Object, param_ptr, &v1);
            world_position_object(Object, &v0, &v0);
            world_position_normal(Object, &v1, &v1);
            VNormalize(v1, v1);
            Polygon.vert[3].x  = v0.x;
            Polygon.vert[3].y  = v0.y;
            Polygon.vert[3].z  = v0.z;
            Polygon.vert[3].nx = v1.x;
            Polygon.vert[3].ny = v1.y;
            Polygon.vert[3].nz = v1.z;

            Polygon.n = 4;
            scan_convert(&Polygon, &WS);
            }
         }
      }
}

/* Find a sphere that bounds all of the control points of a Bezier patch.
   The values returned are: the center of the bounding sphere, and the
   square of the radius of the bounding sphere. */
void
Bezier_Bounding_Sphere(Patch, center, radius)
   VECTOR (*Patch)[4][4], *center;
   DBL *radius;
{
   DBL r0, r1, xc=0, yc=0, zc=0;
   DBL x0, y0, z0;
   int i, j;
   for (i=0;i<4;i++) {
      for (j=0;j<4;j++) {
         xc += (*Patch)[i][j].x;
         yc += (*Patch)[i][j].y;
         zc += (*Patch)[i][j].z;
         }
      }
   xc /= 16.0;
   yc /= 16.0;
   zc /= 16.0;
   r0 = 0.0;
   for (i=0;i<4;i++) {
      for (j=0;j<4;j++) {
         x0 = (*Patch)[i][j].x - xc;
         y0 = (*Patch)[i][j].y - yc;
         z0 = (*Patch)[i][j].z - zc;
         r1 = x0 * x0 + y0 * y0 + z0 * z0;
         if (r1 > r0) r0 = r1;
         }
      }
  center->x = xc; center->y = yc; center->z = zc;
  *radius = r0;
}

void
Bezier_Subpatch_Display(Object, Patch, color)
   OBJECT *Object;
   VECTOR (*Patch)[4][4];
   TEXTURE *color;
{
   VECTOR v0, vertices[4];
   DBL xp[4], yp[4];
   DBL plane[4];

   v0 = (*Patch)[0][0]; world_position_object(Object, &v0, &vertices[0]);
   v0 = (*Patch)[0][3]; world_position_object(Object, &v0, &vertices[1]);
   v0 = (*Patch)[3][3]; world_position_object(Object, &v0, &vertices[2]);
   v0 = (*Patch)[3][0]; world_position_object(Object, &v0, &vertices[3]);
   if (!Shade_Flag) {
      project_points(Eye, 4, &vertices[0], &xp[0], &yp[0], NULL);
      show_polygon(color, 1, 4, &xp[0], &yp[0]);
      }
   else {
      polygon_normal(4, &vertices[0], (VECTOR *)&plane[0], &plane[3]);
      /*
      fill_rectangle(&plane[0], color, &vertices[0], &vertices[1],
                     &vertices[2], &vertices[3]);
      */
   }
}

void
Bezier_Split_Left_Right(Patch, Left_Patch, Right_Patch)
   VECTOR (*Patch)[4][4], (*Left_Patch)[4][4], (*Right_Patch)[4][4];
{
   VECTOR Temp1[4], Temp2[4], Half;
   int i, j;
   for (i=0;i<4;i++) {
      Temp1[0] = (*Patch)[i][0];
      VHalf(Temp1[1], (*Patch)[i][0], (*Patch)[i][1]);
      VHalf(Half, (*Patch)[i][1], (*Patch)[i][2]);
      VHalf(Temp1[2], Temp1[1], Half);
      VHalf(Temp2[2], (*Patch)[i][2], (*Patch)[i][3]);
      VHalf(Temp2[1], Half, Temp2[2]);
      VHalf(Temp1[3], Temp1[2], Temp2[1]);
      Temp2[0] = Temp1[3];
      Temp2[3] = (*Patch)[i][3];
      for (j=0;j<4;j++) {
         (*Left_Patch)[i][j]  = Temp1[j];
         (*Right_Patch)[i][j] = Temp2[j];
         }
      }
}

void
Bezier_Split_Up_Down(Patch, Top_Patch, Bottom_Patch)
   VECTOR (*Patch)[4][4], (*Top_Patch)[4][4], (*Bottom_Patch)[4][4];
{
   VECTOR Temp1[4], Temp2[4], Half;
   int i, j;

   for (i=0;i<4;i++) {
      /* Split Left */
      Temp1[0] = (*Patch)[0][i];
      VHalf(Temp1[1], (*Patch)[0][i], (*Patch)[1][i]);
      VHalf(Half, (*Patch)[1][i], (*Patch)[2][i]);
      VHalf(Temp1[2], Temp1[1], Half);
      VHalf(Temp2[2], (*Patch)[2][i], (*Patch)[3][i]);
      VHalf(Temp2[1], Half, Temp2[2]);
      VHalf(Temp1[3], Temp1[2], Temp2[1]);
      Temp2[0] = Temp1[3];
      Temp2[3] = (*Patch)[3][i];
      for (j=0;j<4;j++) {
         (*Bottom_Patch)[j][i] = Temp1[j];
         (*Top_Patch)[j][i]    = Temp2[j];
         }
      }
}

/* See how close to a plane a subpatch is, the patch must have at least
   three distinct vertices. A negative result from this function indicates
   that a degenerate value of some sort was encountered. */
DBL
determine_subpatch_flatness(Patch)
   VECTOR (*Patch)[4][4];
{
   VECTOR vertices[4], n, b0, b1, TempV;
   DBL d, l0, l1, dist, temp1;
   int i, j;

   vertices[0] = (*Patch)[0][0];
   vertices[1] = (*Patch)[0][3];
   VSub(TempV, vertices[0], vertices[1]);
   VLength(temp1, TempV);
   if (fabs(temp1) < EPSILON) {
      /* Degenerate in the V direction for U = 0. This is ok if the other
         two corners are distinct from the lower left corner - I'm sure there
         are cases where the corners coincide and the middle has good values,
         but that is somewhat pathalogical and won't be considered. */
      vertices[1] = (*Patch)[3][3];
      VSub(TempV, vertices[0], vertices[1]);
      VLength(temp1, TempV);
      if (fabs(temp1) < EPSILON) {
         printf("Degenerate 1\n");
         return -1.0;
         }
      vertices[2] = (*Patch)[3][0];
      VSub(TempV, vertices[0], vertices[1]);
      VLength(temp1, TempV);
      if (fabs(temp1) < EPSILON) {
         printf("Degenerate 2\n");
         return -1.0;
         }
      VSub(TempV, vertices[1], vertices[2]);
      VLength(temp1, TempV);
      if (fabs(temp1) < EPSILON) {
         printf("Degenerate 3\n");
         return -1.0;
         }
      }
   else {
      vertices[2] = (*Patch)[3][0];
      VSub(TempV, vertices[0], vertices[1]);
      VLength(temp1, TempV);
      if (fabs(temp1) < EPSILON) {
         vertices[2] = (*Patch)[3][3];
         VSub(TempV, vertices[0], vertices[2]);
         VLength(temp1, TempV);
         if (fabs(temp1) < EPSILON) {
            printf("Degenerate 4\n");
            return -1.0;
            }
         VSub(TempV, vertices[1], vertices[2]);
         VLength(temp1, TempV);
         if (fabs(temp1) < EPSILON) {
            printf("Degenerate 5\n");
            return -1.0;
            }
         }
      else {
         VSub(TempV, vertices[1], vertices[2]);
         VLength(temp1, TempV);
         if (fabs(temp1) < EPSILON) {
            printf("Degenerate 6\n");
            return -1.0;
            }
         }
      }
   /* Now that a good set of candidate points has been found, find the
      plane equations for the patch */
   if (determine_plane_equations(3, &vertices[0], &n, &d, &l0, &l0, &b0, &b1)) {
      /* Step through all vertices and see what the maximum distance from the
         plane happens to be. */
      dist = 0.0;
      for (i=0;i<4;i++) {
         for (j=0;j<4;j++) {
            temp1 = fabs(point_plane_distance(&((*Patch)[i][j]), &n, &d));
            if (temp1 > dist) dist = temp1;
            }
         }
      return dist;
      }
   else {
      printf("Failed to determine plane equations\n");
      return -1.0;
      }
}

int
flat_enough(Object, Patch)
   OBJECT *Object;
   VECTOR (*Patch)[4][4];
{
   VECTOR Center;
   DBL Dist;
   Dist = determine_subpatch_flatness(Patch);
   if (Dist < 0.0)
      return 0;
   else if (Dist < ((BEZIER *)Object)->Bezier_Params->Division_Epsilon)
      return 1;
   else
      return 0;
}

int
flat_enough0(Object, Patch)
   OBJECT *Object;
   VECTOR (*Patch)[4][4];
{
   VECTOR Center;
   DBL Radius;
   Bezier_Bounding_Sphere(Patch, &Center, &Radius);
   if (Radius < ((BEZIER *)Object)->Bezier_Params->Division_Epsilon)
      return 1;
   else
      return 0;
}

void
Bezier_Subdivider(Object, Patch, color, Depth)
   OBJECT *Object;
   VECTOR (*Patch)[4][4];
   TEXTURE *color;
   int Depth;
{
   VECTOR Lower_Left[4][4], Lower_Right[4][4];
   VECTOR Upper_Left[4][4], Upper_Right[4][4];

   /* U -> Left/Right, V -> Up/Down */

   if (Depth >= Object->Parameters->U_Steps)
      if (Depth >= Object->Parameters->V_Steps)
         Bezier_Subpatch_Display(Object, Patch, color);
      else {
         Bezier_Split_Up_Down(Patch, &Lower_Left, &Upper_Left);
         if (flat_enough(Object, &Lower_Left))
            Bezier_Subpatch_Display(Object, &Lower_Left, color);
         else
            Bezier_Subdivider(Object, &Lower_Left, color, Depth+1);
         if (flat_enough(Object, &Upper_Left))
            Bezier_Subpatch_Display(Object, &Upper_Left, color);
         else
            Bezier_Subdivider(Object, &Upper_Left, color, Depth+1);
         }
   else if (Depth >= Object->Parameters->V_Steps) {
      Bezier_Split_Left_Right(Patch, &Lower_Left, &Lower_Right);
      if (flat_enough(Object, &Lower_Left))
         Bezier_Subpatch_Display(Object, &Lower_Left, color);
      else
         Bezier_Subdivider(Object, &Lower_Left, color, Depth+1);
      if (flat_enough(Object, &Lower_Right))
         Bezier_Subpatch_Display(Object, &Lower_Right, color);
      else
         Bezier_Subdivider(Object, &Lower_Right, color, Depth+1);
      }
   else {
      Bezier_Split_Left_Right(Patch, &Lower_Left, &Lower_Right);
      Bezier_Split_Up_Down(&Lower_Left, &Lower_Left, &Upper_Left);
      Bezier_Split_Up_Down(&Lower_Right, &Lower_Right, &Upper_Right);
      if (flat_enough(Object, &Lower_Left))
         Bezier_Subpatch_Display(Object, &Lower_Left, color);
      else
         Bezier_Subdivider(Object, &Lower_Left, color, Depth+1);
      if (flat_enough(Object, &Upper_Left))
         Bezier_Subpatch_Display(Object, &Upper_Left, color);
      else
         Bezier_Subdivider(Object, &Upper_Left, color, Depth+1);
      if (flat_enough(Object, &Lower_Right))
         Bezier_Subpatch_Display(Object, &Lower_Right, color);
      else
         Bezier_Subdivider(Object, &Lower_Right, color, Depth+1);
      if (flat_enough(Object, &Upper_Right))
         Bezier_Subpatch_Display(Object, &Upper_Right, color);
      else
         Bezier_Subdivider(Object, &Upper_Right, color, Depth+1);
      }
}

void
Bezier_Displayer(Object, color)
   OBJECT *Object;
   TEXTURE *color;
{
   Bezier_Subdivider(Object, ((BEZIER *)Object)->Bezier_Params->Patch,
                     color, 0);
}

void
Radial_Displayer(Object, color)
   OBJECT *Object;
   TEXTURE *color;
{
   RADIAL *Radial = (RADIAL *)Object;
   long offset;
   int i, j, k;
   DBL xp[4], yp[4];
   DBL u, v, x, y, z, param[2], *param_ptr;
   VECTOR v0, v1;
   DBL u_min = Object->Parameters->Min_U;
   DBL u_max = Object->Parameters->Max_U;
   DBL v_min = Object->Parameters->Min_V;
   DBL v_max = Object->Parameters->Max_V;
   int u_steps = Object->Parameters->U_Steps;
   int v_steps = Object->Parameters->V_Steps;
   DBL u_range, v_range, delta_u, delta_v;
   VECTOR vertices[4];
   int uind, vind, ucnt, vcnt, val, poly[128];
   DBL plane[4];
   int strip_count, l;
   char outstuff[80];

   param_ptr = &param[0];

   u_range = u_max - u_min;
   v_range = v_max - v_min;

   /* Calculate the grid and project it. */
   delta_u = u_range / (DBL)u_steps;
   delta_v = v_range / (DBL)v_steps;
      /* Set the draw color */
      setcolor(WHITE);

      /* Apply lighting to each sub sector in the radial. */
      for (i=0,u=u_min;i<u_steps;i++,u+=delta_u) {
         for (j=0,v=v_min;j<v_steps;j++,v+=delta_v) {
            param[0] = u; param[1] = v;
            Evaluate(Object, param_ptr, &v0);
            world_position_object(Object, &v0, &vertices[0]);
            param[1] = v+delta_v;
            Evaluate(Object, param_ptr, &v0);
            world_position_object(Object, &v0, &vertices[1]);
            param[0] = u+delta_u;
            Evaluate(Object, param_ptr, &v0);
            world_position_object(Object, &v0, &vertices[2]);
            param[1] = v;
            Evaluate(Object, param_ptr, &v0);
            world_position_object(Object, &v0, &vertices[3]);

            project_points(Eye, 4, &vertices[0], &xp[0], &yp[0], NULL);
            for (k=0;k<8;k+=2) {
               poly[k] = xp[k/2];
               poly[k+1] = yp[k/2];
               convert(poly[k], poly[k+1]);
               }
            /* Sample the points to see what color it should be. */
            ucnt = (int)((DBL)Default_Sector_Count /
                         (DBL)Object->Parameters->U_Steps);
            vcnt = (int)((DBL)Default_Radial_Count /
                         (DBL)Object->Parameters->V_Steps);
            uind = i * ucnt;
            vind = j * vcnt;
            if (uind < 0 || uind > Default_Sector_Count) {
               printf("Bad u val: %d\n", uind);
               val = 0;
               }
            else if (vind < 0 || vind > Default_Radial_Count) {
               printf("Bad v val: %d\n", vind);
               val = 0;
               }
            else {
               val = 0;
               for (k=0;k<ucnt;k++)
                  for (l=0;l<vcnt;l++) {
                     val += Radial->Radial_Params->plat[vind+l][uind+k];
                     }
               /* Normalize val to the number of available fill patterns. */
               val = (int)(4.0 * (DBL)val / (DBL)(ucnt * vcnt));
               }
#ifdef TURBOC
            if (val == 0)
               setfillstyle(EMPTY_FILL, BLACK);
            else if (val == 1)
               setfillstyle(XHATCH_FILL, WHITE);
            else if (val == 2)
               setfillstyle(CLOSE_DOT_FILL, WHITE);
            else
               setfillstyle(SOLID_FILL, WHITE);
#endif
            if (Radial->Radial_Params->Display_Type == 0 || val > 0)
               fillpoly(4, poly);
            }
         }
}

void
Polygon_Displayer(Object, color)
   OBJECT *Object;
   TEXTURE *color;
{
   VECTOR vertices[MAX_SIDES_PER_POLYGON];
   DBL xp[MAX_SIDES_PER_POLYGON], yp[MAX_SIDES_PER_POLYGON];
   POLYGON *Polygon = (POLYGON *)Object;
   int i, cnt;

   cnt = Polygon->Polygon_Params->Vertex_Count;

   for (i=0;i<cnt;i++)
      world_position_object(Object, &Polygon->Polygon_Params->Vertices[i],
                            &vertices[i]);

   /* Perform perspective transformation on the points. */
   project_points(Eye, cnt, &vertices[0],
                  &xp[0], &yp[0], NULL);
   /* Take the projected points and display them */
   show_polygon(color, Polygon->Polygon_Params->Close_Flag,
                Polygon->Polygon_Params->Vertex_Count,
                &xp[0], &yp[0]);
}

void
Generic_Rotater1(Object, Angles)
   OBJECT *Object;
   VECTOR *Angles;
{
   /* Rotate the entire figure about the world origin */
   Generic_Rotater2(Object, Angles);
   rotate_points(Angles, 1, &(Object->Orientation.Origin),
                 &(Object->Orientation.Origin));
}

void
Generic_Rotater2(Object, Angles)
   OBJECT *Object;
   VECTOR *Angles;
{
   /* Rotate the figure about its own origin */
   rotate_points(Angles, 1, &(Object->Orientation.Major_Axis),
                 &(Object->Orientation.Major_Axis));
   rotate_points(Angles, 1, &(Object->Orientation.Minor_Axis1),
                 &(Object->Orientation.Minor_Axis1));
   rotate_points(Angles, 1, &(Object->Orientation.Minor_Axis2),
                 &(Object->Orientation.Minor_Axis2));
}

/* Rotate about the viewpoint. */
void
Viewpoint_Rotater1(Object, Angles)
   OBJECT *Object;
   VECTOR *Angles;
{
   Generic_Rotater1(Object, Angles);
   Normalize_View(&WS);
}

void
Generic_Translater(Object, Offset)
   OBJECT *Object;
   VECTOR *Offset;
{
   VAdd(Object->Orientation.Origin, Object->Orientation.Origin, *Offset);
}

void
Generic_Scaler(Object, Factors)
   OBJECT *Object;
   VECTOR *Factors;
{
   Object->Orientation.Scale.x *= Factors->x;
   Object->Orientation.Scale.y *= Factors->y;
   Object->Orientation.Scale.z *= Factors->z;
}

void
Polygon_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   /* No evaluation is done. */;
}

void
Bezier_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   bezier_value(Result, param[0], param[1],
                ((BEZIER *)Object)->Bezier_Params->Patch);
}

void
Radial_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   *Result = Object->Orientation.Major_Axis;
}

void
Radial_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL theta, dist;
   DBL len = ((RADIAL *)Object)->Radial_Params->Sector_Length;

   theta = (TWO_PI * param[1]);
   dist = len * param[0];

   Result->x = dist * cos(theta);
   Result->y = dist * sin(theta);
   Result->z = 0.0;
}

void
Ellipse_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL x, y, z, theta, phi;

   theta = (TWO_PI * param[0] - PI) / 2.0;
   phi   = TWO_PI * param[1];

   Result->x = cos(phi) * cos(theta);
   Result->y = sin(phi) * cos(theta);
   Result->z = sin(theta);
}

void
Hyperboloid_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL x, y, z, theta, phi;

   theta = TWO_PI * param[1];
   phi   = 1.76274717 * (0.5 - param[0]);

   Result->x = cos(theta) * cosh(phi);
   Result->y = sin(theta) * cosh(phi);
   Result->z = sinh(phi);
}

void
Paraboloid_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL x, y, z, theta, phi;

   theta = TWO_PI * param[0];
   phi   = param[1];

   Result->x = cos(theta) * phi;
   Result->y = sin(theta) * phi;
   Result->z = phi * phi;
}

void
Cone_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL x, y, z, theta, phi;

   theta = TWO_PI * param[0];
   phi   = param[1];

   Result->x = cos(theta) * phi;
   Result->y = sin(theta) * phi;
   Result->z = phi;
}

void
Hyperbolic_Paraboloid_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL x, y, z, theta, phi;

   theta = TWO_PI * param[0] - PI;
   phi   = param[1];

   Result->x = cosh(theta) * phi;
   Result->y = sinh(theta) * phi;
   Result->z = phi * phi;
}

void
Coil_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL x, y, z, theta, phi;
   DBL a, c, h, k;
   DBL u, v;

   u = param[1];
   v = param[0];

   a = ((COIL *)Object)->Coil_Params->a;
   c = ((COIL *)Object)->Coil_Params->c; /* Vertical displacement per 2 PI */
   k = ((COIL *)Object)->Coil_Params->k;

   theta = TWO_PI * u;
   phi   = TWO_PI * v;

   Result->x = (a * cos(theta)) + c * phi;
   Result->y = ((k + a * sin(theta)) * cos(phi));
   Result->z = ((k + a * sin(theta)) * sin(phi));
}

void
Squishy_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL x, y, z, theta, phi;
   DBL a, b, c, h, k;

   a = ((SQUISHY *)Object)->Squishy_Params->a;
   b = ((SQUISHY *)Object)->Squishy_Params->b;
   c = ((SQUISHY *)Object)->Squishy_Params->c;

   theta = TWO_PI * param[0];
   phi   = TWO_PI * param[1];

   Result->x = (a + c * cos(theta)) * cos(phi);
   Result->y = (b + c * cos(theta)) * sin(phi);
   Result->z = theta;
}

void
Plane_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   Result->x = param[0];
   Result->y = param[1];
   Result->z = 0.0;
}

void
Torus_Evaluater(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL x, y, z, theta, phi;
   DBL a, k;

   a = ((TORUS *)Object)->Torus_Params->a;
   k = ((TORUS *)Object)->Torus_Params->k;

   theta = TWO_PI * param[1];
   phi   = TWO_PI * param[0];

   Result->x = a * cos(theta);
   Result->y = ((k + a * sin(theta)) * cos(phi));
   Result->z = ((k + a * sin(theta)) * sin(phi));
}

void
Bezier_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   bezier_normal(Result, param[0], param[1],
                 ((BEZIER *)Object)->Bezier_Params->Patch);
}

void
Ellipse_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL theta = (TWO_PI * param[0] - PI) / 2.0;
   DBL phi   = TWO_PI * param[1];
   Result->x = cos(theta) * cos(phi);
   Result->y = cos(theta) * sin(phi);
   Result->z = sin(theta);
}

void
Hyperboloid_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL theta, phi;
   VECTOR vc, v0, v1;
   DBL len;

   theta = TWO_PI * param[1];
   phi   = 1.76274717 * (0.5 - param[0]);

   v0.x = -1.0 * sin(theta) * cosh(phi);
   v1.x =        cos(theta) * sinh(phi);

   v0.y =  1.0 * cos(theta) * cosh(phi);
   v1.y =        sin(theta) * sinh(phi);

   v0.z = 0.0;
   v1.z = cosh(phi);

   VCross(vc, v0, v1);
   VLength(len, vc);
   VInverseScale(vc, vc, len);
   *Result = vc;
}

void
Paraboloid_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL theta, phi;
   VECTOR vc, v0, v1;
   DBL len;

   theta = TWO_PI * param[0];
   phi   = param[1];

   v0.x = -sin(theta) * phi;
   v1.x = cos(theta);

   v0.y = cos(theta) * phi;
   v1.y = sin(theta);

   v0.z = 0.0;
   v1.z = 2.0 * phi;

   VCross(vc, v0, v1);
   len = sqrt(vc.x * vc.x + vc.y * vc.y + vc.z * vc.z);
   if (fabs(len) < 0.0001) {
      Result->x =  0.0;
      Result->y =  0.0;
      Result->z = -1.0;
      }
   else {
      Result->x = vc.x / len;
      Result->y = vc.y / len;
      Result->z = vc.z / len;
      }
}

void
Cone_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL theta, phi;
   VECTOR vc, v0, v1;
   DBL len;

   theta = TWO_PI * param[0];
   phi   = param[1];

   v0.x = -sin(theta) * phi;
   v1.x = cos(theta);

   v0.y = cos(theta) * phi;
   v1.y = sin(theta);

   v0.z = 0.0;
   v1.z = 1.0;

   VCross(vc, v0, v1);
   len = sqrt(vc.x * vc.x + vc.y * vc.y + vc.z * vc.z);
   if (fabs(len) < 0.0001) {
      Result->x =  0.0;
      Result->y =  0.0;
      Result->z = -1.0;
      }
   else {
      VInverseScale(vc, vc, len);
      }
}

void
Hyperbolic_Paraboloid_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL theta, phi;
   VECTOR vc, v0, v1;
   DBL len;

   theta = TWO_PI * param[0] - PI;
   phi   = param[1];

   v0.x = -1.0 * phi * (exp(theta) - exp(-theta)) / 2.0;
   v1.x = cosh(theta);

   v0.y = phi * (exp(theta) + exp(-theta)) / 2.0;
   v1.y = sinh(theta);

   v0.z = 0.0;
   v1.z = 2.0 * phi;

   VCross(vc, v0, v1);
   VLength(len, vc);
   VInverseScale(vc, vc, len);
   *Result = vc;
}

void
Coil_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL theta, phi;
   VECTOR v0, v1, vc, vd;
   DBL len;
   DBL a, c, h, k;
   DBL u, v;

   u = param[1];
   v = param[0];

   a = ((COIL *)Object)->Coil_Params->a;
   c = ((COIL *)Object)->Coil_Params->c;
   k = ((COIL *)Object)->Coil_Params->k;

   theta = TWO_PI * u;
   phi   = TWO_PI * v;

   /* Center of the coil */
   VSet(v0, c * phi, k * cos(phi), k * sin(phi));
   /* Point on the coil */
   VSet(v1, (a * cos(theta)) + c * phi,
            (k + a * sin(theta)) * cos(phi),
            (k + a * sin(theta)) * sin(phi));
   /* Direction from center to point */
   VSub(vd, v1, v0);
   VNormalize(vd, vd);

   v0.x = -a*sin(theta);
   v1.x = c;

   v0.y = a*cos(theta)*cos(phi);
   v1.y = (-a*sin(theta)-k)*sin(phi);

   v0.z = a*cos(theta)*sin(phi);
   v1.z = (a*sin(theta)+k)*cos(phi);

   VCross(vc, v0, v1);
   VLength(len, vc);
   VInverseScale(vc, vc, len);
   VDot(len, vd, vc);
   if (len < 0.0) {
      VScale(vc, vc, -1.0);
      }
   *Result = vc;
}

void
Squishy_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL theta, phi;
   DBL a, b, c, h, k;
   VECTOR v0, v1, vd;
   DBL len;

   a = ((SQUISHY *)Object)->Squishy_Params->a;
   b = ((SQUISHY *)Object)->Squishy_Params->b;
   c = ((SQUISHY *)Object)->Squishy_Params->c;

   theta = TWO_PI * param[0];
   phi   = TWO_PI * param[1];

   VSet(v0, (a + c * cos(theta)) * cos(phi),
            (b + c * cos(theta)) * sin(phi),
            theta);
   VSet(v1, cos(phi), sin(phi), theta);
   VSub(vd, v0, v1);

   v0.x = -c * sin(theta) * cos(phi);
   v1.x = (-c * sin(theta) - a) * sin(phi);

   v0.y = -c * sin(theta) * sin(phi);
   v1.y = (c * cos(theta) + b) * cos(phi);

   v0.z = 1.0;
   v1.z = 0.0;

   VCross(*Result, v0, v1);
   VLength(len, *Result);
   if (len == 0.0) {
      printf("Failed normal in squishy_normaler\n");
      VSet(*Result, 1.0, 0.0, 0.0);
      }
   else
      VInverseScale(*Result, *Result, len);
   VDot(len, *Result, vd);
   if (len < 0.0)
      VScale(*Result, *Result, -1.0);
}

void
Plane_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   DBL len;
   Result->x = Object->Orientation.Major_Axis.x;
   Result->y = Object->Orientation.Major_Axis.y;
   Result->z = Object->Orientation.Major_Axis.z;
   VLength(len, *Result);
   VInverseScale(*Result, *Result, len);
}

void
Polygon_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   Result->x = 1.0;
   Result->y = 0.0;
   Result->z = 0.0;
}

void
Torus_Normaler(Object, param, Result)
   OBJECT *Object;
   DBL *param;
   VECTOR *Result;
{
   VECTOR vc, v0, v1;
   DBL x, y, z, theta, phi;
   DBL a, b, h, k, len;

   a = ((TORUS *)Object)->Torus_Params->a;
   k = ((TORUS *)Object)->Torus_Params->k;

   theta = TWO_PI * param[1];
   phi   = TWO_PI * param[0];

   VSet(v0, -a*sin(theta), a*cos(theta)*cos(phi), a*cos(theta)*sin(phi));
   VSet(v1, 0.0, -(k+a*sin(theta))*sin(phi), (k+a*sin(theta))*cos(phi));
   VCross(vc, v0, v1);
   VDot(len, vc, vc);
   len = sqrt(len);
   if (fabs(len) < 0.0001) {
      printf("Bad length in Torus Normaler\n");
      Result->x = 1.0; Result->y = 0.0; Result->z = 0.0;
      }
   else {
      Result->x = vc.x / len;
      Result->y = vc.y / len;
      Result->z = vc.z / len;
      }
}
